from PyQt5 import QtOpenGL
from OpenGL.GL import *


class GLCanvas(QtOpenGL.QGLWidget):

    def __init__(self, _controller=[], _view=[]):
        super(GLCanvas, self).__init__()

        # canvas sizes and model clipping window limits
        self.width = 0   # width of canvas (horizontal raster size)
        self.height = 0  # height of canvas (vertical raster size)
        self.left = -10.0    # left limit of clipping window (world)
        self.right = 10.0    # right limit of clipping window (world)
        self.bottom = -10.0  # bottom limit of clipping window (world)
        self.top = 10.0      # top limit of clipping window (world)

        # canvas aggregation objects
        self.controller = _controller  # handle to app controller
        self.view = _view  # handle to app view

        # OpenGL display lists and display update flags
        self.viewDsp = 0  # GL list index for model display
        self.updatedDsp = False  # if true, model display is updated

        # factors for moving (pan) and zooming (in or out) clipping
        # window (world) with respect to current window sizes
        self.panFac = 0.1
        self.zoomFac = 0.1

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # --------------------GENERAL PROPERTY METHODS-------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def setController(self, _controller):
        self.controller = _controller

    # ---------------------------------------------------------------------
    def getController(self):
        return self.controller

    # ---------------------------------------------------------------------
    def setView(self, _view):
        self.view = _view

    # ---------------------------------------------------------------------
    def getView(self):
        return self.view

    # ---------------------------------------------------------------------
    def resetViewDisplay(self):
        if self.view is None:
            return
        self.updatedDsp = False
        self.update()

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # --------------------CANVAS PREDEFINED SLOTS--------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def initializeGL(self):
        # set canvas background color
        color = self.view.getColorBackground()
        glClearColor(color[0], color[1], color[2], 1.0)
        glClear(GL_COLOR_BUFFER_BIT)

        # set some graphics primitives attributes
        glEnable(GL_LINE_SMOOTH)
        glLineWidth(0.5)
        glPointSize(6.0)

    # ---------------------------------------------------------------------
    def resizeGL(self, _width, _height):

        # Avoid division by zero
        if _width == 0:
            _width = 1

        # store GL canvas sizes in object properties
        self.width = _width
        self.height = _height

        # Setup the viewport to canvas dimesions
        glViewport(0, 0, self.width, self.height)

        # reset the coordinate system
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        # Establish the clipping window in world model coordinates
        # based on model bounding box, and set up an orthographic projection
        self.fitWorldToViewport()

    # ---------------------------------------------------------------------
    def paintGL(self):

        # clear the buffer with the current color
        glClear(GL_COLOR_BUFFER_BIT)

        # Display view (if there is one), storing it in a display list
        if (self.view is not None) and not(self.view.isEmpty()):
            if not self.updatedDsp:
                if self.viewDsp > 0:
                    glDeleteLists(self.viewDsp, 1)
                self.viewDsp = self.makeDisplayView()

            if self.viewDsp > 0:
                glCallList(self.viewDsp)
                self.updatedDsp = True

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # -----------------------DISPLAY FUNCTIONS-----------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def drawPatches(self):
        patches = self.view.getPatches()
        for ptch in patches:
            if self.view.isPatchSelected(ptch):
                color = self.view.getColorPatchSelection()
            else:
                color = self.view.getColorPatch()
            glColor3f(color[0], color[1], color[2])
            pts = self.view.getPatchPts(ptch)
            if len(pts) == 3:
                glBegin(GL_TRIANGLES)
                glVertex2f(pts[0].getX(), pts[0].getY())
                glVertex2f(pts[1].getX(), pts[1].getY())
                glVertex2f(pts[2].getX(), pts[2].getY())
                glEnd()
            else:
                glBegin(GL_POLYGON)
                for j in range(len(pts)):
                    glVertex2f(pts[j].getX(), pts[j].getY())
                glEnd()

    # ---------------------------------------------------------------------
    def drawSegments(self):
        segments = self.view.getSegments()
        for seg in segments:
            if self.view.isSegmentSelected(seg):
                color = self.view.getColorSegmentSelection()
            else:
                color = self.view.getColorSegment()
            glColor3f(color[0], color[1], color[2])
            pts = self.view.getSegmentPts(seg)
            glBegin(GL_LINE_STRIP)
            for j in range(len(pts)):
                glVertex2f(pts[j].getX(), pts[j].getY())
            glEnd()

    # ---------------------------------------------------------------------
    def drawPoints(self):
        points = self.view.getPoints()
        glBegin(GL_POINTS)
        for pnt in points:
            if self.view.isPointSelected(pnt):
                color = self.view.getColorPointSelection()
            else:
                color = self.view.getColorPoint()
            glColor3f(color[0], color[1], color[2])
            pt = self.view.getPointLoc(pnt)
            glVertex2f(pt.getX(), pt.getY())
        glEnd()

    # ---------------------------------------------------------------------
    def makeDisplayView(self):
        if (self.view is None) or (self.view.isEmpty()):
            return 0

        list = glGenLists(1)
        glNewList(list, GL_COMPILE)
        self.drawPatches()
        self.drawSegments()
        self.drawPoints()
        glEndList()
        return list

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # -----------FUNCTIONS TO MANAGE VISUALIZATION WINDOW LIMITS-----------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def scaleWorldWindow(self, _scaleFac):
        # Compute canvas viewport aspect ratio.
        vpr = self.height / self.width

        # Get current window center.
        # **** COMPLETE HERE: GLCANVAS_01 ****

        # Set new window sizes based on scaling factor
        # **** COMPLETE HERE: GLCANVAS_02 ****

        # Adjust window to keep the same aspect ratio of the viewport
        # **** COMPLETE HERE: GLCANVAS_03 ****

        # Establish the clipping volume by setting up an
        # orthographic projection
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(self.left, self.right, self.bottom, self.top, -1.0, 1.0)

        # Setup display model in model coordinates
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        # Force redisplay
        self.update()

    # ---------------------------------------------------------------------
    def panWorldWindow(self, _panX, _panY):
        # Shift current window
        # **** COMPLETE HERE: GLCANVAS_04 ****

        # Establish the clipping volume by setting up an
        # orthographic projection
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(self.left, self.right, self.bottom, self.top, -1.0, 1.0)

        # Setup display model in model coordinates
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        # Force redisplay
        self.update()

    # ---------------------------------------------------------------------
    def fitWorldToViewport(self):
        # Setup world space window limits based on model bounding box
        if (self.view is None) and (self.view.isEmpty()):
            self.scaleWorldWindow(1.0)
        else:
            self.left, self.right, self.bottom, self.top = self.view.getBoundBox()
            scaleFac = 1.0 + self.zoomFac
            self.scaleWorldWindow(scaleFac)

    # ---------------------------------------------------------------------
    def zoomIn(self):
        if self.view is None:
            return

        # Based on zoomFac property, scale down clipping window
        # **** COMPLETE HERE: GLCANVAS_05 ****

    # ---------------------------------------------------------------------
    def zoomOut(self):
        if self.view is None:
            return

        # Based on zoomFac property, scale up clipping window
        # **** COMPLETE HERE: GLCANVAS_06 ****

    # ---------------------------------------------------------------------
    def panLeft(self):
        if self.view is None:
            return

        # Based on panFac property, compute pan (move) shift in
        # horizontal direction and call method to move window left
        # **** COMPLETE HERE: GLCANVAS_07 ****

    # ---------------------------------------------------------------------
    def panRight(self):
        if self.view is None:
            return

        # Based on panFac property, compute pan (move) shift in
        # horizontal direction and call method to move window right
        # **** COMPLETE HERE: GLCANVAS_08 ****

    # ---------------------------------------------------------------------
    def panDown(self):
        if self.view is None:
            return

        # Based on panFac property, compute pan (move) shift in
        # vertical direction and call method to move window down
        # **** COMPLETE HERE: GLCANVAS_09 ****

    # ---------------------------------------------------------------------
    def panUp(self):
        if self.view is None:
            return

        # Based on panFac property, compute pan (move) shift in
        # vertical direction and call method to move window up
        # **** COMPLETE HERE: GLCANVAS_10 ****

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ----------------------MOUSE EVENT SLOTS------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def wheelEvent(self, event):
        if event.angleDelta().y() > 0:
            self.zoomIn()
        else:
            self.zoomOut()
